#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2014 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import clag.clagthread
    import clag.datasync
    import clag.util as util
    import hashlib
    import ipaddr
    import json
    import nlmanager.nlmanager
    import nlmanager.nlpacket
    import Queue
    import socket
    import struct
    import threading
    import xml.etree.ElementTree as ET
except ImportError, e:
    raise ImportError(str(e) + "- required module not found")

class neighsync(clag.datasync.datasync):
    """
    This class handles everything required to synchronize the Neighbor
    Database, between two CLAG peers.
    """

    # Hardware work queue operation codes.
    _NEIGH_ADD = 1
    _NEIGH_DELETE = 2

    # Neigh Work Queue operation codes
    _WORK_QUEUE_ADD = 1
    _WORK_QUEUE_DELETE = 2
    _WORK_QUEUE_NOP = 3


    def __init__(self, Daemon, Intf, Parser, Log, sendQueue, clagnl):
        """
        Constructor: Initialize data structures, threads, etc.
        """
        super(neighsync, self).__init__()

        #
        #   Save the parameters passed in
        #
        self.Daemon = Daemon
        self.Intf = Intf
        self.Parser = Parser
        self.Log = Log
        self.sendQueue = sendQueue
        self.clagnl = clagnl

        #
        #   Neigh Database - Only local (i.e. non offload) and non-permanent
        #   entries on SVIs will be considered for this database.
        #   A neigh is added to that database when it is reachable on at least
        #   one of the two clag peers. A neigh will be owned by both sides when
        #   once it becomes reachable on both clag peer.
        #   (IP, vlanId) : [family MAC localIf ourNeigh peerNeigh]
        #
        #   {                         0             1                2       3     4
        #    ("11.4.1.10", 104) : [AF_INET, "00:00:00:00:00:02", "vlan104", True, True]
        #    ("11.4.1.11", 104) : [AF_INET, "00:00:00:00:00:04", "vlan104", True, False]
        #    ("11.4.1.12", 104) : [AF_INET, "00:00:00:00:00:07", "vlan104", False, True]
        #   }
        #

        self.neighDBLock = threading.Lock()
        self.neighDB = {}

        self.kernDBLock = threading.Lock()
        self.kernDB = {}

        # netlink notification sent to kernel are processed by this queue
        self.toKernWorkQueue = Queue.Queue()
        util.queueAddNameMap(self.toKernWorkQueue, 'toKernWorkQueue')

        # netlink notification received by kernel are processed by this queue
        self.fromKernWorkQueue = Queue.Queue()
        util.queueAddNameMap(self.fromKernWorkQueue, 'fromKernWorkQueue')
        self.clagnl.SetNeighSync(self)

        self.keep_going = True
        self.stopEvent = clag.clagthread.Event()

        self.updateFromKernelThread = threading.Thread(None, self.UpdateFromKernelT,
                                                "UpdateFromKernel")
        self.updateFromKernelThread.daemon = True
        self.updateFromKernelThread.start()

        self.updateToKernelThread = threading.Thread(None, self.UpdateToKernelT,
                                               "UpdateToKernel")
        self.updateToKernelThread.daemon = True
        self.updateToKernelThread.start()

    def __del__(self):
        """
        Destructor:
        """
        self.SignalThreads()
        self.JoinThreads()

    def UpdateFromKernelT(self):
        try:
            self.UpdateFromKernel()
        except:
            self.Daemon.DumpTraceback()

    def UpdateToKernelT(self):
        try:
            self.UpdateToKernel()
        except:
            self.Daemon.DumpTraceback()

    def SignalThreads(self):
        self.keep_going = False
        self.stopEvent.set()
        try:
            self.toKernWorkQueue.put_nowait((None, None, None, None, None, None))
            util.queueSetMaxLen('toKernWorkQueue')
        except Queue.Full:
            pass
        try:
            self.fromKernWorkQueue.put_nowait((neighsync._WORK_QUEUE_NOP, ()))
            util.queueSetMaxLen('fromKernWorkQueue')
        except Queue.Full:
            pass

    def JoinThreads(self):
        self.updateFromKernelThread.join()
        self.updateToKernelThread.join()

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_NEIGH_SYNC, logStr)

    def log_verbose(self, logStr):
        self.Log.log_verbose(self.Log._LOG_DEBUG_NEIGH_SYNC, logStr)

    def KernWorkQueuePut(self, element):
        if self.keep_going:
            self.toKernWorkQueue.put(element)
            util.queueSetMaxLen('toKernWorkQueue')

    def GetObjectSize(self):
        neighObj = {}
        self.neighDBLock.acquire()
        neighObj['neighDB'] = util.getSize(self.neighDB)
        self.neighDBLock.release()
        self.kernDBLock.acquire()
        neighObj['kernNeighDB'] = util.getSize(self.kernDB)
        self.kernDBLock.release()
        return neighObj

    def EncodedStr(self):
        """
        Return local information in human-readable form
        """
        neighDb = []
        self.neighDBLock.acquire()
        for (dst, vlanId), [family, mac, lif, ourNeigh, peerNeigh] in self.neighDB.iteritems():
            neighInfo = {}
            neighInfo['dst'] = dst
            neighInfo['mac'] = mac
            neighInfo['dev'] = lif
            neighInfo['vid'] = vlanId
            neighInfo['ourNeigh'] = ourNeigh
            neighInfo['peerNeigh'] = peerNeigh
            neighDb.append(neighInfo)
        self.neighDBLock.release()
        return json.dumps(neighDb, sort_keys=True, indent=4)

    def CreateDBCheckSum(self):
        self.neighDBLock.acquire()
        hlist = []
        sortedDB =  dict(sorted(self.neighDB.items()))
        for (dst,vlanId), [family, mac, lif, ourNeigh, peerNeigh] in sortedDB.iteritems():
            if self.Parser.clagRole == self.Parser._ROLE_PRIMARY:
                if ourNeigh:
                    ourNeigh = "ownedByPrimary"
                if peerNeigh:
                    peerNeigh = "ownedBySecondary"
                hlist.extend([dst, family, mac, vlanId, ourNeigh, peerNeigh])
            else:
                if ourNeigh:
                    ourNeigh = "ownedBySecondary"
                if peerNeigh:
                    peerNeigh = "ownedByPrimary"
                hlist.extend([dst, family, mac, vlanId, peerNeigh, ourNeigh])
        self.neighDBLock.release()
        return hashlib.md5(''.join(str(e) for e in hlist)).hexdigest()

    def SyncOnIfaceUp(self, iface, vlanAware):
        self.neighDBLock.acquire()
        if vlanAware:
            vid = self.Intf.GetVlanIdOfSubIf(iface)
        else:
            brPeerVlanMap = self.Intf.GetBridgePeerVlanMap()
            vid = brPeerVlanMap.get(iface, None)
        self.log_verbose("%s up, syncing neighs reachable on peer" % iface)
        for (dst, vlanId), [family, mac, lif, ourNeigh, peerNeigh] in self.neighDB.iteritems():
            if iface == lif and vid == vlanId and peerNeigh:
                self.KernWorkQueuePut((neighsync._NEIGH_ADD, dst,
                                     family, mac, vlanId, lif))
        self.neighDBLock.release()

    def CollectLocalInfo(self, nlm):
        """
        This function is called on a periodic interval to gather up any
        information on the local switch which needs to be collected prior to
        sending that information to the peer switch.
        """
        pass

    def EncodeXmlInfo(self, parent):
        """
        Encodes the local information as XML for sending to the peer on a
        periodic basis.
        """
        if self.Parser.syncDoneToPeer:
            return

        self.neighDBLock.acquire()
        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        xmlEntries = 0
        for (dst, vlanId), [family, mac, lif, ourNeigh, peerNeigh] in self.neighDB.iteritems():
            self.AddXmlEntry(root, "sync", dst, family, mac, vlanId)
            xmlEntries += 1
            if xmlEntries >= 2000 and self.keep_going:
                xmlStr = ET.tostring(root)
                self.sendQueue.put(xmlStr)
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0

        self.neighDBLock.release()
        if xmlEntries and self.keep_going:
            xmlStr = ET.tostring(root)
            self.sendQueue.put(xmlStr)
        self.Log.log("Initial neigh sync to peer done.")


    def PrintElement(self, node, level):
        self.log_verbose("%s%s: %s: %s" % (" " * (level*3), node.tag, node.text,
                                           str(node.attrib)))
        for child in node.findall('*'):
            self.PrintElement(child, level+1)

    def GetWorkQueueDelayed(self, delay=0.001):
        """
        Returns the next entry on the neigh work queue, waiting up to the
        specified delay (default 1 millisecond) for an entry.
        """
        queueEmpty = False
        try:
            (opcode, params) = self.fromKernWorkQueue.get(True, delay)
        except Queue.Empty:
            opcode = neighsync._WORK_QUEUE_NOP
            params = ()
            queueEmpty = True
        return (opcode, params, queueEmpty)

    def AddXmlEntry(self, element, op, dst, family, mac, vlanId):
        """
        Adds an Xml entry to the root element
        """
        ET.SubElement(element, "neigh", {"op" : op, "dst" : str(dst),
                                         "mac" : mac, "vlanId" : str(vlanId),
                                         "family" : str(family)})

    def ClearPeerInfo(self):
        """
        Remove the peer ownership flag from all the entries in the neigh DB
        """
        self.neighDBLock.acquire()
        for (dst, vlanId), [family, mac, lif, ourNeigh, peerNeigh] in self.neighDB.iteritems():
            self.neighDB[(dst, vlanId)] = [family, mac, lif, ourNeigh, False]
        # TODO, if not ourNeigh, delete neigh, should not be needed but just in case
        self.neighDBLock.release()

    def GetVlanIdFromValidEntry(self, k_dst, k_if):
        """
        This function checks if the neigh entry is valid. If not, do not process
        it. If it is a valid entry, it returns the vlan ID.
        """
        k_vlanId = None
        if not k_if:
            return
        if self.Intf.isSubIfName(k_if):
            k_vlanId = self.Intf.GetVlanIdOfSubIf(k_if)
            k_master = self.Intf.GetMasterOfSubIf(k_if)
            if not k_master:
                self.log_verbose("Ignoring neigh %s because interface %s "
                               "has no master" % (k_dst, k_if))
                return
            if not self.Intf.isBridgeName(k_master):
                self.log_verbose("Ignoring neigh %s because master "
                               "of %s is not a bridge" % (k_dst, k_if))
                return
            if not self.Intf.isVlanFilteringBridge(k_master):
                self.log_verbose("Ignoring neigh %s because master of %s is not "
                               "a vlan aware bridge" % (k_dst, k_if))
                return
        elif self.Intf.isBridgeName(k_if) and not self.Intf.isVlanFilteringBridge(k_if):
            brPeerVlanMap = self.Intf.GetBridgePeerVlanMap()
            k_vlanId = brPeerVlanMap.get(k_if, None)
        else:
            self.log_verbose("Ignoring neigh %s dev %s" % (k_dst, k_if))
            return
        return k_vlanId

    def SetLocalOwnershipFlag(self, element, params):
        # this neigh is reachable now, we now own it
        (dst, family, mac, lif, vlanId) = params
        self.neighDBLock.acquire()
        if (dst, vlanId) not in self.neighDB:
            self.log_debug("Adding neighbor %s mac %s dev %s in neighDB, now "
                           "reachable" % (dst, mac, lif))
            self.neighDB[(dst, vlanId)] = [family, mac, lif, True, False]
            self.AddXmlEntry(element, "add", dst, family, mac, vlanId)
        else:
            c_mac = self.neighDB[(dst, vlanId)][1]
            c_if = self.neighDB[(dst, vlanId)][2]
            # is it a different mac/iface/vid now?
            if mac != c_mac:
                self.log_debug("Updating neighbor %s mac %s dev %s in neighDB"
                               % (dst, mac, lif))
                self.neighDB[(dst, vlanId)][1] = mac
            elif lif != c_if:
                # only lif changed
                self.neighDB[(dst, vlanId)][2] = lif
            self.AddXmlEntry(element, "add", dst, family, mac, vlanId)
            if self.neighDB[(dst, vlanId)][3] != True:
                self.log_debug("Neighbor %s mac %s dev %s is now reachable, "
                               "updating neighDB" % (dst, mac, lif))
                self.neighDB[(dst, vlanId)][3] = True
        self.neighDBLock.release()

    def ClearLocalOwnershipFlag(self, element, params):
        # no longer reachable, we don't own this neigh anymore
        (dst, family, mac, lif, vlanId) = params
        self.neighDBLock.acquire()
        if (dst, vlanId) in self.neighDB:
            c_mac = self.neighDB[(dst, vlanId)][1]
            if self.neighDB[(dst, vlanId)][4] == True:
                if mac and mac != c_mac:
                    # mac changed and neigh no longer reachable
                    self.neighDB[(dst, vlanId)][3] = False
                elif self.neighDB[(dst, vlanId)][3] != False:
                    # owned by peer so try to re-add
                    operState = self.Intf.GetOperState(lif)
                    self.log_debug("Neighbor %s mac %s dev %s no longer reachable, "
                                   "updating neighDB and try to re-add" % (dst, c_mac, lif))
                    self.neighDB[(dst, vlanId)][3] = False
                    if operState == "up":
                        self.KernWorkQueuePut((neighsync._NEIGH_ADD, dst, family,
                                               c_mac, vlanId, lif))
            else:
                # not owned by peer so remove the entry
                self.log_debug("Deleting neighbor %s mac %s dev %s in neighDB"
                               % (dst, c_mac, lif))
                self.neighDB.pop((dst, vlanId), None)
            self.AddXmlEntry(element, "delete", dst, family, c_mac, vlanId)
        self.neighDBLock.release()

    def SetPeerOwnershipFlag(self, params):
        (dst, family, mac, lif, vlanId) = params
        operState = self.Intf.GetOperState(lif)
        self.neighDBLock.acquire()
        if (dst, vlanId) not in self.neighDB:
            self.log_debug("Adding neighbor %s mac %s dev %s in neighDB, reachable "
                           "on peer" % (dst, mac, lif))
            self.neighDB[(dst, vlanId)] = [family, mac, lif, False, True]
        else:
            c_mac = self.neighDB[(dst, vlanId)][1]
            # is it a different mac or vid now?
            if mac != c_mac:
                self.log_debug("Updating neighbor %s mac %s dev %s in neighDB"
                               % (dst, mac, lif))
                self.neighDB[(dst, vlanId)][1] = mac
            if self.neighDB[(dst, vlanId)][4] != True:
                self.log_debug("Neighbor %s mac %s dev %s now reachable on "
                               "peer, updating neighDB" % (dst, mac, lif))
                self.neighDB[(dst, vlanId)][4] = True
        if (self.neighDB[(dst, vlanId)][3] != True or mac != c_mac) and operState == "up":
            self.KernWorkQueuePut((neighsync._NEIGH_ADD, dst, family, mac, vlanId, lif))

        self.neighDBLock.release()

    def ClearPeerOwnershipFlag(self, params):
        (dst, family, mac, lif, vlanId) = params
        self.neighDBLock.acquire()
        if (dst, vlanId) in self.neighDB:
            c_mac = self.neighDB[(dst, vlanId)][1]
            if self.neighDB[(dst, vlanId)][3] == True:
                if self.neighDB[(dst, vlanId)][4] != False:
                    self.log_debug("Neighbor %s mac %s dev %s not reachable on "
                                   "peer, updating neighDB" % (dst, mac, lif))
                    self.neighDB[(dst, vlanId)][4] = False
            else:
                # we don't even own it so remove the entry
                self.log_debug("Deleting neighbor %s mac %s dev %s from neighDB"
                               % (dst, c_mac, lif))
                self.neighDB.pop((dst, vlanId), None)
        self.neighDBLock.release()

    def UpdateKernelDB(self, opcode, params):
        (k_dst, k_family, k_mac, k_idx, k_state, k_flags) = params
        self.kernDBLock.acquire()
        if opcode == neighsync._WORK_QUEUE_ADD:
            self.kernDB[(k_dst, k_idx)] = [k_family, k_mac, k_state, k_flags]
        else:
            self.kernDB.pop((k_dst, k_idx), None)
        self.kernDBLock.release()

    def UpdateFromKernel(self):
        """
        This is an independent thread of the script. It is responsible for
        processing entries on the NeighWorkQueue, which may result in
        modifications to the neighDB and messages sent to the peer.
        """
        self.log_verbose("Beginning execution of the thread UpdateFromKernel")
        NUD_REACHABLE = 0x02
        NUD_NOARP = 0x40
        NUD_PERMANENT = 0x80
        NTF_EXT_LEARNED = 0x10
        self.clagnl.DumpNeighs()

        while self.keep_going:
            self.log_verbose("Waiting for an event before continuing execution "
                             "of the thread UpdateFromKernel")
            (opcode, k_params) = self.fromKernWorkQueue.get()

            if self.keep_going:
                queueEmpty = False
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0
                while not queueEmpty and self.keep_going:
                    if opcode == neighsync._WORK_QUEUE_NOP:
                        (opcode, k_params, queueEmpty) = self.GetWorkQueueDelayed()
                        continue

                    self.log_debug("Processing neighbor entry: %d %s"
                                   % (opcode, str(k_params)))
                    (k_dst, k_family, k_mac, k_idx, k_state, k_flags) = k_params

                    k_if  = self.Intf.GetIfName(k_idx)
                    k_vlanId = self.GetVlanIdFromValidEntry(k_dst, k_if)
                    if not k_vlanId:
                        (opcode, k_params, queueEmpty) = self.GetWorkQueueDelayed()
                        continue

                    self.kernDBLock.acquire()
                    if (k_dst, k_idx) in self.kernDB:
                        [family, mac, state, flags] = self.kernDB[(k_dst, k_idx)]
                    self.kernDBLock.release()
                    self.UpdateKernelDB(opcode, k_params)

                    neigh_add = None
                    if opcode == neighsync._WORK_QUEUE_ADD:
                        if (k_dst, k_idx) not in self.kernDB:
                            (opcode, k_params, queueEmpty) = self.GetWorkQueueDelayed()
                            continue

                        if k_state & (NUD_PERMANENT|NUD_NOARP) or k_flags & (NTF_EXT_LEARNED):
                            # remove from neighDB
                            neigh_add = None
                        elif k_state & NUD_REACHABLE:
                            # now reachable, need to add to neighDB
                            neigh_add = True

                    params = [k_dst, k_family, k_mac, k_if, k_vlanId]

                    if neigh_add:
                        self.SetLocalOwnershipFlag(root, params)
                        xmlEntries += 1
                    else:
                        self.ClearLocalOwnershipFlag(root, params)
                        xmlEntries += 1

                    if xmlEntries >= 2000 and self.keep_going:
                        xmlStr = ET.tostring(root)
                        self.sendQueue.put(xmlStr)
                        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                        xmlEntries = 0

                    (opcode, k_params, queueEmpty) = self.GetWorkQueueDelayed()

                if xmlEntries and self.keep_going:
                    #self.PrintElement(root, 0)
                    xmlStr = ET.tostring(root)
                    self.sendQueue.put(xmlStr)
        self.log_verbose("Finished execution of the thread UpdateFromKernel")

    def GetLocalIfBasedOnPeerVlanId(self, vlanId):
        brPeerVlanMap = self.Intf.GetBridgePeerVlanMap()
        for bif, vid in brPeerVlanMap.iteritems():
            if vid == vlanId:
                # assuming vlan unaware bridge
                return bif
        # assuming vlan aware bridge
        return self.Intf.GetSubIfBasedOnVlanId(vlanId)

    def DecodeXmlInfo(self, root):
        """
        Decodes XML-encoded information sent from the peer.
        root - The root element of the XML tree
        """
        self.log_verbose("Beginning DecodeXmlInfo")

        for neigh in root.findall('neigh'):
            r_op = neigh.get('op')
            r_dst = neigh.get('dst')
            r_mac = neigh.get('mac')
            r_vlanId = neigh.get('vlanId')
            r_family = neigh.get('family')
            if r_family == "None":
                r_family = None
            if r_family is not None:
                r_family = int(r_family)
            if r_vlanId == "None":
                r_vlanId = None
            if r_vlanId is not None:
                r_vlanId = int(r_vlanId)
            l_if = self.GetLocalIfBasedOnPeerVlanId(r_vlanId)

            params = [r_dst, r_family, r_mac, l_if, r_vlanId]
            if r_op in ["add", "sync"]:
                self.SetPeerOwnershipFlag(params)
            elif r_op == "delete":
                self.ClearPeerOwnershipFlag(params)
        self.log_verbose("Ending DecodeXmlInfo")

    def DualIfChange(self, ifs, enable):
        """
        Called when an interface becomes or is no longer dual connected
        """
        pass

    def UpdateToKernel(self):
        """
        This function is an independent thread of the script. It compares the
        neigh data base received from the peer with the neighs
        we have already programmed into the hardware. Any new neighs are added.
        """

        neigh_cmd = {
            neighsync._NEIGH_ADD : (nlmanager.nlpacket.RTM_NEWNEIGH,
                                    nlmanager.nlpacket.NLM_F_CREATE | nlmanager.nlpacket.NLM_F_REPLACE,
                                    "added to"),
        }

        self.log_verbose("Beginning execution of the thread UpdateToKernel")
        nlm = neigh_nlmanager(self.Log)

        while self.keep_going:
            self.log_verbose("Waiting for an event before continuing execution of the thread UpdateToKernel")
            (op, dst, family, mac, vlanId, lif) = self.toKernWorkQueue.get()

            if self.keep_going:
                queueEmpty = False
                params = []
                while not queueEmpty and self.keep_going:
                    (rtm_command, nlm_flags, neigh_str) = neigh_cmd.get(op, (None, None, None))
                    if rtm_command is not None:
                        intf = self.Intf.GetIfIndex(lif)
                        if not intf:
                            try:
                                (op, dst, family, mac, vlanId, lif) = self.toKernWorkQueue.get(True, 0.001)
                                intf = self.Intf.GetIfIndex(lif)
                            except Queue.Empty:
                                queueEmpty = True
                            continue
                        params += [(rtm_command, nlm_flags, dst, mac, intf, family)]
                        # TODO Move this to verbose
                        self.log_debug("Neighbor %s %s has been %s %s in VLAN %d"
                                       % (dst, mac, neigh_str, lif, vlanId))
                    else:
                        self.Log.log_error("Unknown op code in work queue: %s" % (op,))

                    try:
                        (op, dst, family, mac, vlanId, lif) = self.toKernWorkQueue.get(True, 0.001)
                    except Queue.Empty:
                        queueEmpty = True

                nlm.neigh_batch(params)

        self.log_verbose("Finished execution of the thread UpdateToKernel")

class neigh_nlmanager(nlmanager.nlmanager.NetlinkManager):

    def __init__(self, Log):
        super(neigh_nlmanager, self).__init__(util.CLAG_NLM_NEIGH_ID)
        self.Log = Log

    def tx_or_concat_message(self, total_message, neigh):
        """
        Concat several of neighs together before TXing.
        """
        PACKET_CONCAT_SIZE = 16384

        if not total_message:
            total_message = neigh.message
        else:
            total_message += neigh.message

        if len(total_message) >= PACKET_CONCAT_SIZE:
            self.tx_nlpacket_raw(total_message)
            total_message = None

        return total_message

    def neigh_batch(self, neighs):

        total_message = None

        for (rtm_command, nlm_flags, dst, mac, intf, family) in neighs:
            debugFlags = self.Log._LOG_DEBUG_NETLINK | self.Log._LOG_DEBUG_NEIGH_SYNC
            debug = (self.Log.getDebug() & debugFlags) == debugFlags
            neigh = nlmanager.nlpacket.Neighbor(rtm_command, debug, self.Log.logger,
                                                self.Log.getHandlerStr() != "syslog")
            neigh.flags = nlmanager.nlpacket.NLM_F_REQUEST | nlm_flags
            neigh.family = family

            if family == socket.AF_INET:
                dst = ipaddr.IPv4Address(dst)
            else:
                dst = ipaddr.IPv6Address(dst)

            neigh.body = struct.pack(neigh.PACK, family, intf, neigh.NUD_STALE, 0, 0)
            neigh.add_attribute(neigh.NDA_LLADDR, mac)
            neigh.add_attribute(neigh.NDA_DST, dst)

            neigh.build_message(self.sequence.next(), self.pid)
            total_message = self.tx_or_concat_message(total_message, neigh)

        if total_message:
            self.tx_nlpacket_raw(total_message)
